@@ -11,6 +11,7 @@ from .types import *
 from .colors import Colors
 from random import randint
 from contextlib import contextmanager
+from pathlib import Path
 
 
 def noop(x):
@@ -30,8 +31,10 @@ def data_threshold(data_type: str) -> int:
 class Connector:
     __last_time_stamp: float = -1
     __last_sub_time: float = 0
-    __record_data__: bool = False
+    __record_data: bool = False
     data: dict[str, dict[str, list[ClientMsg]]] = DictX({})
+    __current_data_frame: dict[str, DataFrame] = DictX({})
+    __latest_data: DataFrame = default_data_frame()
     __devices = {'time_stamp': time_s(), 'devices': []}
     device: Optional[Device] = None
     __server_url: str
@@ -81,6 +84,7 @@ class Connector:
     on_sprite_collision: Union[Callable[[SpriteCollisionMsg], None],
                                Callable[[SpriteCollisionMsg, Connector], None]] = noop
     on_border_overlap: Union[Callable[[BorderOverlapMsg], None], Callable[[BorderOverlapMsg, Connector], None]] = noop
+    on_sprite_clicked: Union[Callable[[SpriteClickedMsg], None], Callable[[SpriteClickedMsg, Connector], None]] = noop
 
     __on_notify_subscribers: Union[Callable[[], None], Callable[[
         DataFrame], None], Callable[[DataFrame, Connector], None]] = noop
@@ -118,6 +122,7 @@ class Connector:
     def __init__(self, server_url: str, device_id: str):
         self.__server_url = server_url
         self.__device_id = device_id
+        self.__current_data_frame[device_id] = default_data_frame()
         self.sio.on('connect', self.__on_connect)
         self.sio.on('disconnect', self.__on_disconnect)
         self.sio.on(SocketEvents.NEW_DATA, self.__on_new_data)
@@ -477,7 +482,8 @@ class Connector:
         for dev_id in dev_ids:
             data_types = self.data[dev_id].keys() if data_type is None else [data_type]
             for dtype in data_types:
-                all_data.extend(self.data[dev_id][dtype])
+                if dtype in self.data[dev_id]:
+                    all_data.extend(self.data[dev_id][dtype])
         all_data.sort(key=lambda d: d['time_stamp'])
         return all_data
 
@@ -537,35 +543,67 @@ class Connector:
         DataMsg, None
             when no data is found, None is returned
         '''
+        if device_id is None:
+            device_id = self.device_id
+        if device_id == '__ALL_DEVICES__':
+            raw = deepcopy(self.__latest_data)
+        else:
+            raw = deepcopy(self.__current_data_frame[device_id])
 
-        data = self.all_data(data_type=data_type, device_id=device_id)
-        if (len(data) == 0):
-            return None
-        return data[-1]
+        if raw is None:
+            if data_type is None:
+                return default('key')
+            return default(data_type)
+
+        if data_type is None:
+            srted = sorted(raw.values(), reverse=True, key=lambda x: x['time_stamp'])
+            return srted[0]
+
+        if data_type in raw:
+            return raw[data_type]
+        return default(data_type)
 
     def latest_pointer(self, device_id: str = '__ALL_DEVICES__') -> Union[ColorPointer, GridPointer, None]:
-        return self.latest_data('pointer', device_id=device_id)
+        return self.latest_data(device_id=device_id, data_type='pointer')
 
     def latest_color_pointer(self, device_id: str = '__ALL_DEVICES__') -> Union[ColorPointer, None]:
-        data = self.all_data(data_type='pointer', device_id=device_id)
-        for pkg in reversed(data):
-            if pkg.context == 'color':
-                return pkg
+        return self.latest_data(device_id=device_id, data_type='color_pointer')
 
     def latest_grid_pointer(self, device_id: str = '__ALL_DEVICES__') -> Union[GridPointer, None]:
-        data = self.all_data(data_type='pointer', device_id=device_id)
-        for pkg in reversed(data):
-            if pkg.context == 'grid':
-                return pkg
+        return self.latest_data(device_id=device_id, data_type='color_pointer')
 
     def latest_gyro(self, device_id: str = '__ALL_DEVICES__') -> Union[None, GyroMsg]:
-        return self.latest_data('gyro', device_id=device_id)
+        return self.latest_data(device_id=device_id, data_type=DataType.GYRO)
 
     def latest_acceleration(self, device_id: str = '__ALL_DEVICES__') -> Union[None, AccMsg]:
-        return self.latest_data('acceleration', device_id=device_id)
+        return self.latest_data(device_id=device_id, data_type=DataType.ACCELERATION)
 
     def latest_key(self, device_id: str = '__ALL_DEVICES__') -> Union[None, KeyMsg]:
-        return self.latest_data('key', device_id=device_id)
+        return self.latest_data(device_id=device_id, data_type=DataType.KEY)
+
+    @property
+    def pointer(self) -> PointerDataMsg:
+        return self.latest_pointer(device_id=self.device_id)
+
+    @property
+    def color_pointer(self) -> ColorPointerMsg:
+        return self.latest_color_pointer(device_id=self.device_id)
+
+    @property
+    def grid_pointer(self) -> GridPointerMsg:
+        return self.latest_grid_pointer(device_id=self.device_id)
+
+    @property
+    def gyro(self) -> GyroMsg:
+        return self.latest_gyro(device_id=self.device_id)
+
+    @property
+    def acceleration(self) -> AccMsg:
+        return self.latest_acceleration(device_id=self.device_id)
+
+    @property
+    def key(self) -> KeyMsg:
+        return self.latest_key(device_id=self.device_id)
 
     def configure_playground(self,
                              width: Optional[Number] = None,
@@ -573,15 +611,42 @@ class Connector:
                              shift_x: Optional[Number] = None,
                              shift_y: Optional[Number] = None,
                              color: Optional[Union[Colors, str]] = None,
+                             images: Optional[Union[Path, str]] = None,
                              **delivery_opts):
+        raw_images = []
+        if images is not None:
+            images = Path(images)
+            if images.is_absolute():
+                pass
+            else:
+                rpath = Path.cwd().rglob(str(images))
+                try:
+                    while True:
+                        new_p = next(rpath)
+                        if new_p.is_dir():
+                            images = new_p
+                            break
+                except:
+                    pass
+
+            if not images.is_dir():
+                raise f'Image path {images} not found'
+            for image in images.iterdir():
+                if image.suffix in ['.jpg', '.jpeg', '.png', '.svg']:
+                    raw = image.read_bytes()
+                    name = image.stem
+                    file_type = image.suffix
+                    raw_images.append({'name': name, 'image': raw, 'type': file_type[1:]})
+
         config = {
             'type': DataType.PLAYGROUND_CONFIG,
             'config': without_none({
                 'width': width,
                 'height': height,
-                'shift_x': shift_x,
-                'shift_y': shift_y,
-                'color': color
+                'shift_x': shift_x if shift_x is not None else 0,
+                'shift_y': shift_y if shift_y is not None else 0,
+                'color': color,
+                'images': raw_images
             })
         }
         self.emit(SocketEvents.NEW_DATA, config, **delivery_opts)
@@ -604,8 +669,10 @@ class Connector:
                 reset_time: Optional[Number] = None,
                 speed: Optional[Number] = None,
                 text: Optional[str] = None,
+                image: Optional[str] = None,
                 time_span: Optional[Number] = None,
                 width: Optional[Number] = None,
+                rotate: Optional[Number] = None,
                 **delivery_opts):
             '''
             Optional
@@ -655,6 +722,12 @@ class Connector:
             text : str
                 the text that is displayed on the sprite.
 
+            image : str
+                name of the image to be displayed. The image must be set when the playground is configured. No file ending expected.
+
+            rotate : Number
+                degrees to rotate the sprite clockwise 
+
             time_span : Number
                 the time a sprite lives
             '''
@@ -672,6 +745,8 @@ class Connector:
                 'reset_time': reset_time,
                 'speed': speed,
                 'text': text,
+                'image': image,
+                'rotate': rotate,
                 'time_span': time_span,
                 'width': width
             }
@@ -699,7 +774,7 @@ class Connector:
 
     def add_sprite(
             self,
-            id: Optional[str],
+            id: Optional[str] = None,
             clickable: Optional[bool] = None,
             collision_detection: Optional[bool] = None,
             color: Optional[Union[Colors, str]] = None,
@@ -712,8 +787,10 @@ class Connector:
             reset_time: Optional[Number] = None,
             speed: Optional[Number] = None,
             text: Optional[str] = None,
+            image: Optional[str] = None,
             time_span: Optional[Number] = None,
             width: Optional[Number] = None,
+            rotate: Optional[Number] = None,
             **delivery_opts):
         '''
         Optional
@@ -763,6 +840,12 @@ class Connector:
         text : str
             the text that is displayed on the sprite.
 
+        image : str
+            name of the image to be displayed. The image must be set when the playground is configured. No file ending expected.
+
+        rotate : Number
+            degrees to rotate the sprite clockwise
+
         time_span : Number
             the time a sprite lives
         '''
@@ -780,6 +863,8 @@ class Connector:
             'reset_time': reset_time,
             'speed': speed,
             'text': text,
+            'rotate': rotate,
+            'image': image,
             'time_span': time_span,
             'width': width
         }
@@ -834,26 +919,41 @@ class Connector:
             return deepcopy(grid)
         return [deepcopy(grid)]
 
-    def set_grid_at(self, row: int, column: int, color: CssColorType, base_color: Optional[BaseColor] = None, **delivery_opts):
+    def update_cell(self, row: Optional[int] = None, column: Optional[int] = None, color: Optional[CssColorType] = None, cellNumber: Optional[int] = None, base_color: Optional[BaseColor] = None, enumerate: bool = None, **delivery_opts):
+        '''
+        sets the color of the current grid at the given row and column
+        '''
+        self.update_grid(row=row, column=column, color=color, cellNumber=cellNumber,
+                         base_color=base_color, enumerate=enumerate, **delivery_opts)
+
+    def set_grid_at(self, row: Optional[int] = None, column: Optional[int] = None, color: Optional[CssColorType] = None, cellNumber: Optional[int] = None, base_color: Optional[BaseColor] = None, enumerate: bool = None, **delivery_opts):
+        '''
+        sets the color of the current grid at the given row and column
+        '''
+        self.update_grid(row=row, column=column, color=color, cellNumber=cellNumber,
+                         base_color=base_color, enumerate=enumerate, **delivery_opts)
+
+    def update_grid(self, row: Optional[int] = None, column: Optional[int] = None, color: Optional[CssColorType] = None, cellNumber: Optional[int] = None, base_color: Optional[BaseColor] = None, enumerate: bool = None, **delivery_opts):
         '''
         sets the color of the current grid at the given row and column
         '''
-        self.__set_local_grid_at(row, column, color)
         grid_msg = {
             'type': 'grid_update',
             'row': row,
             'column': column,
+            'number': cellNumber,
             'color': color,
             'base_color': base_color,
+            'enumerate': enumerate,
             'time_stamp': self.current_time_stamp
         }
         self.emit(
             SocketEvents.NEW_DATA,
-            grid_msg,
+            without_none(grid_msg),
             **delivery_opts
         )
 
-    def set_image(self, image: Union[List[str], str], base_color: Union[BaseColor] = None, **delivery_opts):
+    def set_image(self, image: Union[List[str], str], base_color: Union[BaseColor] = None, enumerate: bool = None, **delivery_opts):
         '''
         Parameters
         ----------
@@ -885,7 +985,7 @@ class Connector:
         phone.set_image(image)
         ```
         '''
-        return self.set_grid(image, base_color=base_color, **delivery_opts)
+        return self.set_grid(image, base_color=base_color, enumerate=enumerate, **delivery_opts)
 
     def reset_grid(self, **delivery_opts):
         self.set_grid([['white']], **delivery_opts)
@@ -919,7 +1019,7 @@ class Connector:
 
         self.__last_sent_grid.grid = cast(List[List[CssColorType]], list(raw_grid))
 
-    def set_grid(self, grid: ColorGrid, base_color: Union[BaseColor] = None, **delivery_opts):
+    def set_grid(self, grid: ColorGrid, base_color: BaseColor = None, enumerate: bool = None, **delivery_opts):
         '''
         Parameters
         ----------
@@ -957,11 +1057,12 @@ class Connector:
             'type': 'grid',
             'grid': grid,
             'base_color': base_color,
+            'enumerate': enumerate,
             'time_stamp': self.current_time_stamp
         }
         self.emit(
             SocketEvents.NEW_DATA,
-            grid_msg,
+            without_none(grid_msg),
             **delivery_opts
         )
 
@@ -1157,14 +1258,14 @@ class Connector:
 
     def start_recording(self):
         self.clean_data()
-        self.__record_data__ = True
+        self.__record_data = True
 
     def stop_recording(self):
-        self.__record_data__ = False
+        self.__record_data = False
 
     @property
     def is_recording(self):
-        return self.__record_data__
+        return self.__record_data
 
     def __on_connect(self):
         logging.info('SocketIO connected')
@@ -1190,6 +1291,23 @@ class Connector:
         except Exception:
             pass
 
+    def __update_current_data_frame(self, data: dict):
+        if data['device_id'] not in self.__current_data_frame:
+            self.__current_data_frame[data['device_id']] = default_data_frame()
+        if data['type'] in ['key', 'acceleration', 'gyro']:
+            self.__current_data_frame[data['device_id']][data['type']] = data
+        elif data['type'] == DataType.POINTER:
+            contex = data['context']
+            tkey = f'{contex}_pointer'
+            self.__current_data_frame[data['device_id']][tkey] = data
+
+    def __update_latest_data(self, data: dict):
+        self.__latest_data[data['type']] = data
+        if data['type'] == DataType.POINTER:
+            contex = data['context']
+            tkey = f'{contex}_pointer'
+            self.__latest_data[tkey] = data
+
     def __on_new_data(self, data: dict):
         data = DictX(data)
         if 'device_id' not in data:
@@ -1201,8 +1319,11 @@ class Connector:
         if data['type'] not in self.data[data['device_id']]:
             self.data[data['device_id']][data['type']] = []
 
+        self.__update_current_data_frame(data)
+        self.__update_latest_data(data)
+
         data_len = len(self.data[data['device_id']][data['type']])
-        if not self.__record_data__ and (data_len >= data_threshold(data['type'])):
+        if not self.__record_data and (data_len >= data_threshold(data['type'])):
             self.data[data['device_id']][data['type']].pop(0)
 
         self.data[data['device_id']][data['type']].append(cast(ClientMsg, data))
@@ -1214,6 +1335,9 @@ class Connector:
 
     def __distribute_new_data_callback(self, data: DataMsg):
         if 'type' in data:
+            if data['type'] in [DataType.ACCELERATION, DataType.GYRO]:
+                self.__callback('on_sensor', data)
+
             if data['type'] == DataType.KEY:
                 self.__callback('on_key', data)
                 if data['key'] == 'F1':
@@ -1224,28 +1348,28 @@ class Connector:
                     self.__callback('on_f3', data)
                 elif data['key'] == 'F4':
                     self.__callback('on_f4', data)
-            if data['type'] in [DataType.ACCELERATION, DataType.GYRO]:
-                self.__callback('on_sensor', data)
-            if data['type'] == DataType.ACCELERATION:
+            elif data['type'] == DataType.ACCELERATION:
                 self.__callback('on_acceleration', data)
-            if data['type'] == DataType.GYRO:
+            elif data['type'] == DataType.GYRO:
                 self.__callback('on_gyro', data)
-            if data['type'] == DataType.POINTER:
+            elif data['type'] == DataType.POINTER:
                 self.__callback('on_pointer', data)
-            if data['type'] == DataType.INPUT_RESPONSE:
+            elif data['type'] == DataType.INPUT_RESPONSE:
                 self.__responses.append(cast(InputResponseMsg, data))
-            if data['type'] == DataType.ALERT_CONFIRM:
+            elif data['type'] == DataType.ALERT_CONFIRM:
                 self.__alerts.append(cast(AlertConfirmMsg, data))
-            if data['type'] == DataType.SPRITE_OUT:
+            elif data['type'] == DataType.SPRITE_OUT:
                 sprite = first(lambda s: s.id == data.sprite_id, self.__sprites)
                 if sprite:
                     self.__sprites.remove(sprite)
                 self.__callback('on_sprite_out', data)
-            if data['type'] == DataType.SPRITE_COLLISION:
+            elif data['type'] == DataType.SPRITE_COLLISION:
                 data['sprites'] = list(map(lambda s: DictX(s), data['sprites']))
                 self.__callback('on_sprite_collision', data)
-            if data['type'] == DataType.BORDER_OVERLAP:
+            elif data['type'] == DataType.BORDER_OVERLAP:
                 self.__callback('on_border_overlap', data)
+            elif data['type'] == DataType.SPRITE_CLICKED:
+                self.__callback('on_sprite_clicked', data)
 
         if 'broadcast' in data and data['broadcast'] and self.on_broadcast_data is not None:
             self.__callback('on_broadcast_data', data)
